<?

 /*************************************************************************************
 * Filename: flowchart.php
 *
 * This file is part of the Graphics JAS plugin.
 * Copyright 2003-2016 by Matthijs Schuurman
 *
 * The Graphics JAS plugin is free software under the terms of the
 * GNU General Public License version 3 or any later version.
 *
 * For information about the license please refer to:
 * LICENSE.txt in the root of the plugin or http://jas.junerules.com/plugins/graphics/
 *************************************************************************************/


 require_once($_path["include"]."data.php");
 require_once($_path["include"]."errors.php");
 require_once($_path["include"]."mutilate.php");
 require_once($_path["project"]["plugins"]."graphics/drawing.php");

 class FlowChart extends Drawing
 {
  function __construct()
  {
   parent::__construct();

   $this->default=Array(

    //action shape
    "action" => Array(
     "text" => "action_%d",
     "size" => Array(
      "width" => 100,
      "height" => 50,
     ),
     "margin" => Array(
      "top" => 10,
      "right" => 10,
      "bottom" => 10,
      "left" => 10,
     ),
     "color" => Array(
      "background" => "3CB5F4",
      "text" => "FFFFFF",
      "border" => "878787",
     ),
     "border" => 1,
    ),

    //label shape
    "label" => Array(
     "text" => "label_%d",
     "size" => Array(
      "width" => 100,
      "height" => 25,
     ),
     "margin" => Array(
      "top" => 10,
      "right" => 10,
      "bottom" => 10,
      "left" => 10,
     ),
     "color" => Array(
      "background" => false,
      "text" => "878787",
      "border" => false,
     ),
     "border" => 0,
    ),

    //connector
    "connector" => Array(
     "text" => "connector_%d",
     "type" => "corner",
     "color" => Array(
      "background" => "3CB5F4",
      "text" => "878787",
      "border" => false,
     ),
     "width" => 1,
    ),


   );

   $this->actionCounter=0;
   $this->connectorCounter=0;
   $this->labelCounter=0;
  }


  /*
   *
   * shapes
   */

  //action shape 
  function action($id=false)
  {
   if(!$id) //generate id
    $id=$this->default["action"]["text"];

   //replace %d with action number
   $id=str_replace("%d", Mutilate::leadingZerosS($this->actionCounter+1, 2), $id);

   if(!$this->rectangle($id, "0,0", "0,0"))
    return false;

   $this->items[$id]["flowchart"]=Array(
    "type" => "action",
   );

   //set id as text
   if(!$this->text($id))
   {
    unset($this->items[$id]);
    return false;
   }

   //auto positioning
   if(!$this->setPosition($id)) 
   {
    unset($this->items[$id]);
    return false;
   }

   //default sizing
   if(!$this->setSize($id))
   {
    unset($this->items[$id]);
    return false;
   }

   //default margining
   if(!$this->setMargin($id)) 
   {
    unset($this->items[$id]);
    return false;
   }

   //default action colors
   if(!$this->setColor($id)) 
   {
    unset($this->items[$id]);
    return false;
   }

   //default action border
   if(!$this->setBorder($id)) 
   {
    unset($this->items[$id]);
    return false;
   }

   $this->actionCounter++;   
   return $id;
  }


  //label shape 
  function label($id=false)
  {
   if(!$id) //generate id
    $id=$this->default["label"]["text"];

   //replace %d with action number
   $id=str_replace("%d", Mutilate::leadingZerosS($this->labelCounter+1, 2), $id);

   if(!$this->rectangle($id, "0,0", "0,0"))
    return false;

   $this->items[$id]["flowchart"]=Array(
    "type" => "label",
   );

   //set id as text
   if(!$this->text($id))
   {
    unset($this->items[$id]);
    return false;
   }

   //auto positioning
   if(!$this->setPosition($id)) 
   {
    unset($this->items[$id]);
    return false;
   }

   //default sizing
   if(!$this->setSize($id))
   {
    unset($this->items[$id]);
    return false;
   }

   //default margining
   if(!$this->setMargin($id)) 
   {
    unset($this->items[$id]);
    return false;
   }

   //default label colors
   if(!$this->setColor($id)) 
   {
    unset($this->items[$id]);
    return false;
   }

   //default label border
   if(!$this->setBorder($id)) 
   {
    unset($this->items[$id]);
    return false;
   }

   $this->labelCounter++;   
   return $id;
  }


  /*
   *
   * connectors
   */

  //connector
  function connector($from, $to, $type=false, $id=false)
  {
   if(!isset($this->items[$from]))
   {
    $this->e->add("Can't add connector, ID \"".$from."\" doesn't exist", "error");
    return false;
   }

   if(!isset($this->items[$to]))
   {
    $this->e->add("Can't add connector, ID \"".$to."\" doesn't exist", "error");
    return false;
   }

   if(!$id) //generate id
    $id=$this->default["connector"]["text"];

   //replace %d with action number
   $id=str_replace("%d", Mutilate::leadingZerosS($this->connectorCounter+1, 2), $id);

   if(!$type)
    $type=$this->default["connector"]["type"];

   if($type!="line"&&$type!="corner")
   {
    $this->e->add("Illegal connector type for ID \"".$id."\", use: \"line\" or \"corner\"", "warning");
    $$position=$this->default["connector"]["type"]; //use default connector type
   }

   if($type=="line")
   {
    if(!$this->line($id, "0,0"))
     return false;
   }
   else
   {
    if(!$this->corner($id, "0,0"))
     return false;
   }

   $this->items[$id]["flowchart"]=Array(
    "type" => "connector",
    "from" => $from,
    "to" => $to,
   );

   //auto points determination
   if(!$this->setPoints($id)) 
   {
    unset($this->items[$id]);
    return false;
   }

   //default connector colors
   if(!$this->setColor($id)) 
   {
    unset($this->items[$id]);
    return false;
   }

   //default connector width
   if(!$this->setWidth($id)) 
   {
    unset($this->items[$id]);
    return false;
   }

   $this->connectorCounter++;   
   return $id;
  }


  /*
   *
   * properties
   */


  //position, based on center of the shape
  function setPosition($id, $horizontal=false, $vertical=false)
  {
   if(!isset($this->items[$id]))
   {
    $this->e->add("Can't set position, ID \"".$id."\" doesn't exist", "error");
    return false;
   }

   if($this->items[$id]["flowchart"]["type"]=="connector")
   {
    $this->e->add("Can't set position, ID \"".$id."\" isn't a shape", "error");
    return false;
   }

   if($horizontal===false)
    $horizontal="auto";

   if($vertical===false)
    $vertical="auto";

   $positions=Array("horizontal", "vertical");
   foreach($positions as $position)
   {
    if($$position!="auto"&&!is_numeric($$position))
    {
     $this->e->add("Illegal $position position for ID \"".$id."\", use: \"auto\" or a valid number", "warning");
     $$position="auto"; //use auto
    }
   }

   $this->items[$id]["flowchart"]["position"]=Array(
    "horizontal" => $horizontal,
    "vertical" => $vertical,
   );

   return true;
  }


  //points
  function setPoints($id, $from=false, $to=false)
  {
   if(!isset($this->items[$id]))
   {
    $this->e->add("Can't set points, ID \"".$id."\" doesn't exist", "error");
    return false;
   }

   if($this->items[$id]["flowchart"]["type"]!="connector")
   {
    $this->e->add("Can't set position, ID \"".$id."\" isn't a connector", "error");
    return false;
   }

   if($from===false)
    $from="auto";

   if($to===false)
    $to="auto";

   $points=Array("from", "to");
   $allowedPoints=Array("top", "right", "bottom", "left");
   foreach($points as $point)
   {
    if($$point!="auto"&&!in_array($$point, $allowedPoints))
    {
     $this->e->add("Illegal $point point for ID \"".$id."\", use: \"auto\" or one of the following: \"".implode("\", \"", $allowedPoints)."\"", "warning");
     $$point="auto"; //use auto
    }
   }

   $this->items[$id]["flowchart"]["point"]=Array(
    "from" => $from,
    "to" => $to,
   );

   return true;
  }


  //size
  function setSize($id, $width=false, $height=false)
  {
   if(!isset($this->items[$id]))
   {
    $this->e->add("Can't set size, ID \"".$id."\" doesn't exist", "error");
    return false;
   }

   if($width===false)
    $width="default";

   if($height===false)
    $height="default";

   $sizes=Array("width", "height");
   foreach($sizes as $size)
   {
    if($$size!="default"&&!is_numeric($$size))
    {
     $this->e->add("Illegal $size for ID \"".$id."\", use: \"default\" or a valid number", "warning");
     $$size="default"; //use default
    }
   }

   $this->items[$id]["flowchart"]["size"]=Array(
    "width" => $width,
    "height" => $height,
   );

   return true;
  }


  //margin
  function setMargin($id, $top=false, $right=false, $bottom=false, $left=false)
  {
   if(!isset($this->items[$id]))
   {
    $this->e->add("Can't set margin, ID \"".$id."\" doesn't exist", "error");
    return false;
   }

   if($top===false)
    $top="default";

   if($right===false)
    $right="default";

   if($bottom===false)
    $bottom="default";

   if($left===false)
    $left="default";

   $margins=Array("top", "right", "bottom", "left");
   foreach($margins as $margin)
   {
    if($$margin!="default"&&!is_numeric($$margin))
    {
     $this->e->add("Illegal $margin margin for ID \"".$id."\", use: \"default\" or a valid number", "warning");
     $$margin="default"; //use default
    }
   }

   $this->items[$id]["flowchart"]["margin"]=Array(
    "top" => $top,
    "left" => $left,
    "bottom" => $bottom,
    "right" => $right,
   );

   return true;
  }


  //color
  function setColor($id, $background=false, $text=false, $border=false)
  {
   if(!isset($this->items[$id]))
   {
    $this->e->add("Can't set color, ID \"".$id."\" doesn't exist", "error");
    return false;
   }

   if($background===false)
    $background="default";

   if($text===false)
    $text="default";

   if($border===false)
    $border="default";

   $colors=Array("background", "text", "border");
   foreach($colors as $color)
   {
    if($$color!="default"&&!preg_match("/^[0-F]+$/", $$color))
    {
     $this->e->add("Illegal $color color for ID \"".$id."\", use: \"default\" or a valid (HTML) color", "warning");
     $$color="default"; //use default
    }
   }

   $this->items[$id]["flowchart"]["color"]=Array(
    "background" => $background,
    "text" => $text,
    "border" => $border,
   );

   return true;
  }


  //border
  function setBorder($id, $border=false)
  {
   if(!isset($this->items[$id]))
   {
    $this->e->add("Can't set border size, ID \"".$id."\" doesn't exist", "error");
    return false;
   }

   if($border===false)
    $border="default";

   if($border!="default"&&!is_numeric($border))
   {
    $this->e->add("Illegal border size for ID \"".$id."\", use: \"default\" or a valid number", "warning");
    $border="default"; //use default
   }

   $this->items[$id]["flowchart"]["border"]=$border;

   return true;
  }


  //width
  function setWidth($id, $width=false)
  {
   if(!isset($this->items[$id]))
   {
    $this->e->add("Can't set width, ID \"".$id."\" doesn't exist", "error");
    return false;
   }

   if($width===false)
    $width="default";

   if($width!="default"&&!is_numeric($width))
   {
    $this->e->add("Illegal width for ID \"".$id."\", use: \"default\" or a valid number", "warning");
    $width="default"; //use default
   }

   $this->items[$id]["flowchart"]["width"]=$width;

   return true;
  }


  /*
   *
   * rendering
   */

  function render($width=800, $height=600)
  {
   if(!$this->size($width, $height))
    return false;

   /*
    * box
    * - top:    top position - (height/2 + border + margin top)
    * - left:   left position - (width/2 + border + margin left)
    * - bottom: bottom position + (height/2 + border + margin bottom)
    *   or:     top + margin top + border + height + border + margin bottom
    * - right:  right position - (width/2 + border + margin right)
    *   or:     left + margin left + border + width + border + margin right
    */

   //set default values
   foreach(array_keys($this->items) as $id)
   {
    if(!isset($this->items[$id]["flowchart"])) //skip non flowchart shape/connector
     continue;

    $type=$this->items[$id]["flowchart"]["type"];
    if($type=="connector")
    {
     //width
     $width=($this->items[$id]["flowchart"]["width"]!="default"?$this->items[$id]["flowchart"]["width"]:$this->default[$type]["width"]);
     $this->width($id, $width); //don't bother checking errors, parent::render will display them anyway
    }
    else //actions/labels
    {
     //border
     $border=($this->items[$id]["flowchart"]["border"]!="default"?$this->items[$id]["flowchart"]["border"]:$this->default[$type]["border"]);
     $this->border($id, $border); //don't bother checking errors, parent::render will display them anyway
    }

    //colors
    $background=($this->items[$id]["flowchart"]["color"]["background"]!="default"?$this->items[$id]["flowchart"]["color"]["background"]:$this->default[$type]["color"]["background"]);
    $text=($this->items[$id]["flowchart"]["color"]["text"]!="default"?$this->items[$id]["flowchart"]["color"]["text"]:$this->default[$type]["color"]["text"]);
    $border=($this->items[$id]["flowchart"]["color"]["border"]!="default"?$this->items[$id]["flowchart"]["color"]["border"]:$this->default[$type]["color"]["border"]);
    $this->color($id, $background, $border, $text); //don't bother checking errors, parent::render will display them anyway
   }


   //static shape positioning
   foreach(array_keys($this->items) as $id)
   {
    if(!isset($this->items[$id]["flowchart"])) //skip non flowchart shape/connector
     continue;

    $type=$this->items[$id]["flowchart"]["type"];
    if($type!="connector") //actions/labels
    {
     if($this->items[$id]["flowchart"]["position"]["horizontal"]=="auto"&&$this->items[$id]["flowchart"]["position"]["vertical"]=="auto") //skip auto position shapes
      continue;

     //border size
     $border=($this->items[$id]["flowchart"]["border"]!="default"?$this->items[$id]["flowchart"]["border"]:$this->default[$type]["border"]);

     //set final horizontal position
     if($this->items[$id]["flowchart"]["position"]["horizontal"]!="auto")
     {
      $horizontal=$this->items[$id]["flowchart"]["position"]["horizontal"];

      $width=($this->items[$id]["flowchart"]["size"]["width"]!="default"?$this->items[$id]["flowchart"]["size"]["width"]:$this->default[$type]["size"]["width"]);
      $margin["left"]=($this->items[$id]["flowchart"]["margin"]["left"]!="default"?$this->items[$id]["flowchart"]["margin"]["left"]:$this->default[$type]["margin"]["left"]);
      $margin["right"]=($this->items[$id]["flowchart"]["margin"]["right"]!="default"?$this->items[$id]["flowchart"]["margin"]["right"]:$this->default[$type]["margin"]["right"]);
      
      $left=$horizontal-(($width/2)+$border);
      $right=$border+$left+$width+$border; 
     }
     else
     {
      $left=0;
      $right=0;
     }

     //set final horizontal position
     if($this->items[$id]["flowchart"]["position"]["horizontal"]!="auto")
     {
      $vertical=$this->items[$id]["flowchart"]["position"]["vertical"];

      $height=($this->items[$id]["flowchart"]["size"]["height"]!="default"?$this->items[$id]["flowchart"]["size"]["height"]:$this->default[$type]["size"]["height"]);
      $margin["top"]=($this->items[$id]["flowchart"]["margin"]["top"]!="default"?$this->items[$id]["flowchart"]["margin"]["top"]:$this->default[$type]["margin"]["top"]);
      $margin["bottom"]=($this->items[$id]["flowchart"]["margin"]["bottom"]!="default"?$this->items[$id]["flowchart"]["margin"]["bottom"]:$this->default[$type]["margin"]["bottom"]);

      $top=$vertical-(($height/2)+$border);
      $bottom=$border+$top+$height+$border; 
     }
     else
     {
      $top=0;
      $bottom=0;
     }

     //set new coordinates
     $this->coords($id, $left.",".$top, $right.",".$bottom);
    }
   }


   //auto shape positioning
   foreach(array_keys($this->items) as $id)
   {
    if(!isset($this->items[$id]["flowchart"])) //skip non flowchart shape/connector
     continue;

    $type=$this->items[$id]["flowchart"]["type"];
    if($type!="connector") //actions/labels
    {
     if($this->items[$id]["flowchart"]["position"]["horizontal"]!="auto"&&$this->items[$id]["flowchart"]["position"]["vertical"]!="auto") //skip static position shapes
      continue;

     //border size
     $border=($this->items[$id]["flowchart"]["border"]!="default"?$this->items[$id]["flowchart"]["border"]:$this->default[$type]["border"]);

     //size
     $width=($this->items[$id]["flowchart"]["size"]["width"]!="default"?$this->items[$id]["flowchart"]["size"]["width"]:$this->default[$type]["size"]["width"]);
     $height=($this->items[$id]["flowchart"]["size"]["height"]!="default"?$this->items[$id]["flowchart"]["size"]["height"]:$this->default[$type]["size"]["height"]);

     //margin
     $margin["top"]=($this->items[$id]["flowchart"]["margin"]["top"]!="default"?$this->items[$id]["flowchart"]["margin"]["top"]:$this->default[$type]["margin"]["top"]);
     $margin["right"]=($this->items[$id]["flowchart"]["margin"]["right"]!="default"?$this->items[$id]["flowchart"]["margin"]["right"]:$this->default[$type]["margin"]["right"]);
     $margin["bottom"]=($this->items[$id]["flowchart"]["margin"]["bottom"]!="default"?$this->items[$id]["flowchart"]["margin"]["bottom"]:$this->default[$type]["margin"]["bottom"]);
     $margin["left"]=($this->items[$id]["flowchart"]["margin"]["left"]!="default"?$this->items[$id]["flowchart"]["margin"]["left"]:$this->default[$type]["margin"]["left"]);

     //determine auto positioning based on other "final" positioned shapes
     if($this->items[$id]["flowchart"]["position"]["horizontal"]=="auto"||$this->items[$id]["flowchart"]["position"]["vertical"]=="auto")
     {
      if($this->items[$id]["flowchart"]["position"]["horizontal"]!="auto")
       $horizontal=$this->items[$id]["flowchart"]["position"]["horizontal"];
      else
       $horizontal=(($width/2)+$border+$margin["left"]);

      if($this->items[$id]["flowchart"]["position"]["vertical"]!="auto")
       $vertical=$this->items[$id]["flowchart"]["position"]["vertical"];
      else
       $vertical=(($height/2)+$border+$margin["top"]);

      //make box
      $box["left"]=$horizontal-(($width/2)+$border+$margin["left"]);
      $box["right"]=$margin["left"]+$border+$box["left"]+$width+$border+$margin["right"];
      $box["top"]=$vertical-(($height/2)+$border+$margin["top"]);
      $box["bottom"]=$margin["top"]+$border+$box["top"]+$height+$border+$margin["bottom"];

      $overlaps=true;
      while($overlaps)
      {
       $overlaps=false;
       $skipAutoShapes=false;
       foreach(array_keys($this->items) as $id2) //run through all shapes
       {
        if(!isset($this->items[$id2]["flowchart"])) //skip non flowchart shape/connector
         continue;

        $type=$this->items[$id2]["flowchart"]["type"];
        if($type!="connector") //only shapes
        {
         if($id==$id2) //current shape, all before this one have "final" values, everthing after will determine their own auto position
         {
          $skipAutoShapes=true;
          continue;
         }

         if($skipAutoShapes) //skip non reliable positions
          if($this->items[$id2]["flowchart"]["position"]["horizontal"]=="auto"||$this->items[$id2]["flowchart"]["position"]["vertical"]=="auto")
           continue;

         $box2=$this->getBox($id2, true);

         //detect overlap with shape (partial overlap, fully over or fully inside)
         if(
          (
           ($box["left"]<$box2["left"]||$box["left"]>$box2["right"])&& //box 1 left position not in box 2
           ($box["right"]<$box2["left"]||$box["right"]>$box2["right"])&& //box 1 right position not in box 2
           ($box2["left"]<$box["left"]||$box2["left"]>$box["right"]) //box 2 left position not in box 1
           //implied by previous checks ($box2["right"]<$box["left"]||$box2["right"]>$box["right"]) //box 2 right position not in box 1
          )||(
           ($box["top"]<$box2["top"]||$box["top"]>$box2["bottom"])&& //box 1 top position not in box 2
           ($box["bottom"]<$box2["top"]||$box["bottom"]>$box2["bottom"])&& //box 1 bottom position not in box 2
           ($box2["top"]<$box["top"]||$box2["top"]>$box["bottom"]) //box 2 top position not in box 1
           //implied by previous checks ($box2["bottom"]<$box["top"]||$box2["bottom"]>$box["bottom"]) //box 2 bottom position not in box 1
          )
         )
         continue;

         $overlaps=true;
         //get next position, to the right of the shape
         $horizontal=1+$box2["right"]+$margin["left"]+$border+($width/2);
         $box["left"]=$horizontal-(($width/2)+$border+$margin["left"]);
         $box["right"]=$margin["left"]+$border+$box["left"]+$width+$border+$margin["right"]; 

         if($box["right"]>$this->width) //shape outside of horizontal drawing boundary
         {
          //move down by its own height
          //$vertical+=1+($box2["bottom"]-$box2["top"]);
          $vertical+=$margin["top"]; 
          $box["top"]=$vertical-(($height/2)+$border+$margin["top"]);
          $box["bottom"]=$margin["top"]+$border+$box["top"]+$height+$border+$margin["bottom"];

          //set horizontal position back to left of the drawing
          $horizontal=(($width/2)+$border+$margin["left"]);
          $box["left"]=$horizontal-(($width/2)+$border+$margin["left"]);
          $box["right"]=$margin["left"]+$border+$box["left"]+$width+$border+$margin["right"]; 
         }
        }
       }
      }
     }
     else //no auto positioning, set static values
     {
      $horizontal=$this->items[$id]["flowchart"]["position"]["horizontal"];
      $vertical=$this->items[$id]["flowchart"]["position"]["vertical"];
     }

     //set new coordinates
     $left=$horizontal-(($width/2)+$border);
     $right=$border+$left+$width+$border; 
     $top=$vertical-(($height/2)+$border);
     $bottom=$border+$top+$height+$border; 

     $this->coords($id, $left.",".$top, $right.",".$bottom);
    }
   }


   //connector points determination and routing
   foreach(array_keys($this->items) as $id)
   {
    if(!isset($this->items[$id]["flowchart"])) //skip non flowchart shape/connector
     continue;

    $type=$this->items[$id]["flowchart"]["type"];
    if($type=="connector")
    {
     /*
      * Point determination:
      *     top
      * left * right
      *    bottom
      */

     $i1=$this->items[$this->items[$id]["flowchart"]["from"]];
     $i2=$this->items[$this->items[$id]["flowchart"]["to"]];

     $i1Horizontal=$i1["coords"][0][0]+(($i1["coords"][1][0]-$i1["coords"][0][0])/2); //left 
     $i1Vertical=$i1["coords"][0][1]+(($i1["coords"][1][1]-$i1["coords"][0][1])/2); //top

     $i2Horizontal=$i2["coords"][0][0]+(($i2["coords"][1][0]-$i2["coords"][0][0])/2); //left 
     $i2Vertical=$i2["coords"][0][1]+(($i2["coords"][1][1]-$i2["coords"][0][1])/2); //top

     $i1Margin["top"]=($i1["flowchart"]["margin"]["top"]!="default"?$i1["flowchart"]["margin"]["top"]:$this->default[$i1["flowchart"]["type"]]["margin"]["top"]);
     $i1Margin["right"]=($i1["flowchart"]["margin"]["right"]!="default"?$i1["flowchart"]["margin"]["right"]:$this->default[$i1["flowchart"]["type"]]["margin"]["right"]);
     $i1Margin["bottom"]=($i1["flowchart"]["margin"]["bottom"]!="default"?$i1["flowchart"]["margin"]["bottom"]:$this->default[$i1["flowchart"]["type"]]["margin"]["bottom"]);
     $i1Margin["left"]=($i1["flowchart"]["margin"]["left"]!="default"?$i1["flowchart"]["margin"]["left"]:$this->default[$i1["flowchart"]["type"]]["margin"]["left"]);

     $i2Margin["top"]=($i2["flowchart"]["margin"]["top"]!="default"?$i2["flowchart"]["margin"]["top"]:$this->default[$i2["flowchart"]["type"]]["margin"]["top"]);
     $i2Margin["right"]=($i2["flowchart"]["margin"]["right"]!="default"?$i2["flowchart"]["margin"]["right"]:$this->default[$i2["flowchart"]["type"]]["margin"]["right"]);
     $i2Margin["bottom"]=($i2["flowchart"]["margin"]["bottom"]!="default"?$i2["flowchart"]["margin"]["bottom"]:$this->default[$i2["flowchart"]["type"]]["margin"]["bottom"]);
     $i2Margin["left"]=($i2["flowchart"]["margin"]["left"]!="default"?$i2["flowchart"]["margin"]["left"]:$this->default[$i2["flowchart"]["type"]]["margin"]["left"]);

     $diffHorizontal=$i2Horizontal-$i1Horizontal;
     $diffVertical=$i2Vertical-$i1Vertical;

     if($this->items[$id]["flowchart"]["point"]["from"]=="auto")
     {
      if(
       $diffHorizontal==0|| //horizontally centered
       ($i1["coords"][0][0]<$i2Horizontal&&$i2Horizontal<$i1["coords"][1][0]) //or center of shape 2 is above/under shape 1
      )
      {
       if($diffVertical>0) //shape 1 above 2
        $i1Point="bottom"; 
       else //shape 1 below 2
        $i1Point="top";
      }
      else if(
       $diffVertical==0|| //vertically centered
       ($i2Horizontal<$i1["coords"][0][0]||$i1["coords"][1][0]<$i2Horizontal) //or center of shape 2 is not above/under shape 1
      )
      {
       if($diffHorizontal>0) //shape 1 left of 2
        $i1Point="right";
       else //shape 1 right of 2
        $i1Point="left";
      }
     }
     else
      $i1Point=$this->items[$id]["flowchart"]["point"]["from"];

     if($this->items[$id]["flowchart"]["point"]["to"]=="auto")
     {
      if(
       $diffVertical==0|| //vertically centered
       ($i2["coords"][0][1]<$i1Vertical&&$i1Vertical<$i2["coords"][1][1]) //or center of shape 1 is left/right shape 2
      )
      {
       if($diffHorizontal>0) //shape left
        $i2Point="left";
       else //shape right
        $i2Point="right";
      }
      else if(
       $diffHorizontal==0|| //horizontally centered
       ($i1Vertical<$i2["coords"][0][1]||$i2["coords"][1][1]<$i1Vertical) //or center of shape 1 is not left/right shape 2
      )
      {
       if($diffVertical>0) //shape above
        $i2Point="top";
       else //shape below
        $i2Point="bottom";
      }
     }
     else
      $i2Point=$this->items[$id]["flowchart"]["point"]["to"];

     //make diffs positive values
     if($diffHorizontal<0)
      $diffHorizontal*=-1;
     if($diffVertical<0)
      $diffVertical*=-1;

     //determine actual coordinates
     if($i1Point=="top")
      $start=Array(($i1Horizontal), ($i1["coords"][0][1]-$i1["width"]-1));
     else if($i1Point=="right") 
      $start=Array(($i1["coords"][1][0]+$i1["width"]), ($i1Vertical));
     else if($i1Point=="bottom") 
      $start=Array(($i1Horizontal), (($i1["coords"][1][1]+$i1["width"])));
     else if($i1Point=="left") 
      $start=Array(($i1["coords"][0][0]-$i1["width"]-1), ($i1Vertical));

     if($i2Point=="top")
      $end=Array(($i2Horizontal), ($i2["coords"][0][1]-$i2["width"]-1));
     else if($i2Point=="right") 
      $end=Array(($i2["coords"][1][0]+$i2["width"]), ($i2Vertical));
     else if($i2Point=="bottom") 
      $end=Array(($i2Horizontal), (($i2["coords"][1][1]+$i2["width"])));
     else if($i2Point=="left") 
      $end=Array(($i2["coords"][0][0]-$i2["width"]-1), ($i2Vertical));

     if($this->items[$id]["type"]=="corner") //find path
     {
      $c=0;
      $coords=Array();

      $coords[$c]=$start;
      $c++;

      //store left over x and y
      $xLeft=$end[0]-$start[0];
      $yLeft=$end[1]-$start[1];

      //start may only go in 1 direction, first step is to go out of the margin
      if($i1Point=="top") //must go up
      {
       $coords[$c]=Array($coords[$c-1][0], $coords[$c-1][1]-$i1Margin["top"]/2);
       $direction="up";
       $yLeft+=$i1Margin["top"]/2;
       $c++;
      }
      else if($i1Point=="right") //must go right
      {
       $coords[$c]=Array($coords[$c-1][0]+$i1Margin["right"]/2, $coords[$c-1][1]);
       $direction="right";
       $xLeft-=$i1Margin["right"]/2;
       $c++;
      }
      else if($i1Point=="bottom") //must go down
      {
       $coords[$c]=Array($coords[$c-1][0], $coords[$c-1][1]+$i1Margin["bottom"]/2);
       $direction="down";
       $yLeft-=$i1Margin["bottom"]/2;
       $c++;
      }
      else if($i1Point=="left") //must go left
      {
       $coords[$c]=Array($coords[$c-1][0]-$i1Margin["left"]/2, $coords[$c-1][1]);
       $direction="left";
       $xLeft+=$i1Margin["left"]/2;
       $c++;
      }

      //end may only go in 1 direction, stop at the margin
      if($i2Point=="top") //must go up
       $yLeft-=$i2Margin["top"]/2;
      else if($i2Point=="right") //must go right
       $xLeft+=$i2Margin["right"]/2;
      else if($i2Point=="bottom") //must go down
       $yLeft+=$i2Margin["bottom"]/2;
      else if($i2Point=="left") //must go left
       $xLeft-=$i2Margin["left"]/2;

      while($xLeft!=0||$yLeft!=0)
      {
       //try to progress in current direction
       if(($direction=="up"&&$yLeft<0)||($direction=="down"&&$yLeft>0))
        $move="vertical";
       else if(($direction=="left"&&$xLeft<0)||($direction=="right"&&$xLeft>0))
        $move="horizontal";
       else //no need to go any further, take a turn
       {
        if($direction=="up"||$direction=="down") 
        {
         $move="horizontal";
         if($xLeft==0) //no more horizontal movement 
         {
          if(-10<$yLeft&&$yLeft<10) //pretty much there, XXX: should use i2Margin to see if it's done
           break; //skip to the end

          //go around self
          $divert=($i1["coords"][1][0]-$i1["coords"][0][0])/2;
          $divert+=($i1["flowchart"]["border"]!="default"?$i1["flowchart"]["border"]:$this->default[$i1["flowchart"]["type"]]["border"]);

          $xLeft+=$divert+$i1Margin["right"];
         }
        }
        else //left or right
        {
         $move="vertical";
         if($yLeft==0) //no more vertical movement 
         {
          if(-10<$xLeft&&$xLeft<10) //pretty much there, XXX: should use i2Margin to see if it's done
           break; //skip to the end

          //go around self
          $divert=($i1["coords"][1][1]-$i1["coords"][0][1])/2;
          $divert+=($i1["flowchart"]["border"]!="default"?$i1["flowchart"]["border"]:$this->default[$i1["flowchart"]["type"]]["border"]);

          $yLeft+=$divert+$i1Margin["top"];
         }
        }
       }

       //path detection
       $x=$y=0;
       if($move=="vertical") 
       {
        $y=$yLeft;

        if($ids=$this->detectCollision($coords[$c-1], Array($coords[$c-1][0], $coords[$c-1][1]+$y))) //anticipated move collides with other shapes
        {
         //find closest colliding shape
         $closest=false;
         $closestDistance=false;
         foreach($ids as $id2)
         {
          //get distance 
          $distance=($y>0?($this->items[$id2]["coords"][0][1]-$this->items[$id2]["width"]-1)-$coords[$c-1][1]:($this->items[$id2]["coords"][1][1]+$this->items[$id2]["width"]+1)-$coords[$c-1][1]);
          if($distance<0)
           $distance*=-1;

          //compare to previous lowest
          if($closest===false||$closestDistance>$distance)
          {
           $closest=$id2;
           $closestDistance=$distance;
          }
         }

         if($closest!=$this->items[$id]["flowchart"]["to"]) //near target, don't divert
         {
          //make sure up/down stops before it collides with the object
          $y=$closestDistance; 

          //determine which is the shortest way around
          if(($diffLeff=$this->items[$closest]["coords"][0][0]-$coords[$c-1][0])<0) //get distance to left side
           $diffLeff*=-1;

          if(($diffRight=$this->items[$closest]["coords"][1][0]-$coords[$c-1][0])<0) //get distance to right side
           $diffRight*=-1;

          if($diffLeff>$diffRight) //right side is closer
           $x=$diffRight+$this->items[$closest]["width"]+1; //go right
          else
           $x=-($diffLeff+$this->items[$closest]["width"]+2); //go left

          //move horizontal first, otherwise below its done as second 
          $coords[$c]=Array($coords[$c-1][0]+$x, $coords[$c-1][1]);
          $direction=($x<0?"left":"right");
          $xLeft-=$x;
          $x=0;
          $c++;
         }
        }
       }
       else //horizontal
       {
        $x=$xLeft;

        if($ids=$this->detectCollision($coords[$c-1], Array($coords[$c-1][0]+$x, $coords[$c-1][1]))) //anticipated move collides with other shapes
        {
         //find closest colliding shape
         $closest=false;
         $closestDistance=false;
         foreach($ids as $id2)
         {
          //get distance 
          $distance=($y>0?($this->items[$id2]["coords"][0][0]-$this->items[$id2]["width"]-1)-$coords[$c-1][0]:($this->items[$id2]["coords"][1][0]+$this->items[$id2]["width"]+1)-$coords[$c-1][0]);
          if($distance<0)
           $distance*=-1;

          //compare to previous lowest
          if($closest===false||$closestDistance>$distance)
          {
           $closest=$id2;
           $closestDistance=$distance;
          }
         }

         if($closest!=$this->items[$id]["flowchart"]["to"]) //near target, don't divert
         {
          //make sure left/right stops before it collides with the object
          $x=$closestDistance; 

          //determine which is the shortest way around
          if(($diffTop=$this->items[$closest]["coords"][0][1]-$coords[$c-1][1])<0) //get distance to top
           $diffTop*=-1;

          if(($diffBottom=$this->items[$closest]["coords"][1][1]-$coords[$c-1][1])<0) //get distance to bottom
           $diffBottom*=-1;

          if($diffTop>$diffBottom) //bottom side is closer
           $y=$diffBottom+$this->items[$closest]["width"]+1; //go down
          else
           $y=-($diffTop+$this->items[$closest]["width"]+2); //go up
         }
        }
       }

       if($y) //move vertical
       {
        $coords[$c]=Array($coords[$c-1][0], $coords[$c-1][1]+$y);
        $direction=($y<0?"up":"down");
        $yLeft-=$y;
        $y=0;
        $c++;
       }

       if($x)//move horizontal
       {
        $coords[$c]=Array($coords[$c-1][0]+$x, $coords[$c-1][1]);
        $direction=($x<0?"left":"right");
        $xLeft-=$x;
        $x=0;
        $c++;
       }

       if($c>25) //max 20 steps
        break;
      }

      $coords[$c]=$end;
      $c++;
     }
     else
      $coords=Array($start, $end);

     $this->coords($id, $coords);
    }
   }

   return parent::render();
  }


  function getBox($id, $margin)
  {
   if(!isset($this->items[$id]))
   {
    $this->e->add("Can't get box, ID \"".$id."\" doesn't exist", "error");
    return false;
   }

   if($this->items[$id]["flowchart"]["type"]=="connector")
   {
    $this->e->add("Can't get box, ID \"".$id."\" isn't a shape", "error");
    return false;
   }

   $type=$this->items[$id]["flowchart"]["type"];
   $box["left"]=$this->items[$id]["coords"][0][0];
   $box["right"]=$this->items[$id]["coords"][1][0];
   $box["top"]=$this->items[$id]["coords"][0][1];
   $box["bottom"]=$this->items[$id]["coords"][1][1];

   if($margin)
   {
    $box["left"]-=($this->items[$id]["flowchart"]["margin"]["left"]!="default"?$this->items[$id]["flowchart"]["margin"]["left"]:$this->default[$type]["margin"]["left"]);
    $box["right"]+=($this->items[$id]["flowchart"]["margin"]["right"]!="default"?$this->items[$id]["flowchart"]["margin"]["right"]:$this->default[$type]["margin"]["right"]);
    $box["top"]-=($this->items[$id]["flowchart"]["margin"]["top"]!="default"?$this->items[$id]["flowchart"]["margin"]["top"]:$this->default[$type]["margin"]["top"]);
    $box["bottom"]+=($this->items[$id]["flowchart"]["margin"]["bottom"]!="default"?$this->items[$id]["flowchart"]["margin"]["bottom"]:$this->default[$type]["margin"]["bottom"]);
   }

   return $box;
  }

  function detectCollision($coord1, $coord2) 
  {
   if(!isset($coord1[0])||!isset($coord1[1])||!isset($coord2[0])||!isset($coord2[1]))
   {
    $this->e->add("Can't detect collision, coordinate arrays should contain 2 entries", "error");
    return false;
   }

/*
   //detect method
   if($coord1[0]==$coord2[0]) //x the same
    $method="vertical";
   else if($coord1[1]==$coord2[1]) //y the same
    $method="horizontal";
   else
    $method="box";
*/

   $collisions=Array();
   foreach(array_keys($this->items) as $id) //run through all shapes
   {
    if(!isset($this->items[$id]["flowchart"])) //skip non flowchart shape/connector
     continue;

    $type=$this->items[$id]["flowchart"]["type"];
    if($type!="connector") //only shapes
    {
     $box=$this->getBox($id, false);

     //detect overlap with shape (partial overlap, fully over or fully inside)
     if($coord1[0]<$box["left"]&&$coord2[0]<$box["left"]) //both horizontal coordinates to the left of the box
      continue;

     if($box["right"]<$coord1[0]&&$box["right"]<$coord2[0]) //both horizontal coordinates to the right of the box
      continue;

     if($coord1[1]<$box["top"]&&$coord2[1]<$box["top"]) //both vertical coordinates above the box
      continue;

     if($box["bottom"]<$coord1[1]&&$box["bottom"]<$coord2[1]) //both vertical coordinates below the box
      continue;

     $collisions[]=$id; //colliding ID
    }
   }

   return $collisions;
  }

 }

?>
