<?

 /*************************************************************************************
 * Filename: graph.php
 *
 * This file is part of the Graphics JAS plugin.
 * Copyright 2003-2016 by Matthijs Schuurman
 *
 * The Graphics JAS plugin is free software under the terms of the
 * GNU General Public License version 3 or any later version.
 *
 * For information about the license please refer to:
 * LICENSE.txt in the root of the plugin or http://jas.junerules.com/plugins/graphics/
 *************************************************************************************/


 require_once($_path["include"]."data.php");
 require_once($_path["include"]."errors.php");
 require_once($_path["include"]."mutilate.php");
 require_once($_path["project"]["plugins"]."graphics/drawing.php");

 class Graph extends Drawing
 {
  function __construct()
  {
   parent::__construct();

   $this->autoAddLineLabels=true;
   $this->fontHeight=15;
   $this->fontWidth=8;
   $this->dateFormat="Y-m-d H:i:s";
  }

  function clear($onlyValues=false)
  {
   if($onlyValues)
   {
    foreach(array_keys($this->items) as $id)
     if(isset($this->items[$id]["gtype"]))
      if($this->items[$id]["gtype"]=="line"||$this->items[$id]["gtype"]=="bar")
       unset($this->items[$id]);
   }
   else
   {
    parent::clear();
    $this->border("drawing", 0);
   }
  }

  function axis($id, $orientation, $min=false, $max=false)
  {
   if(isset($this->items[$id]))
   {
    $this->e->add("ID \"".$id."\" already exists", "error");
    return false;
   }

   if($orientation!="horizontal"&&$orientation!="vertical")
   {
    $axe=$orientation;
    $orientation=false;
   }
   else
    $axe=false;

   if($axe)
   {
    if(!isset($this->items[$axe]))
    {
     $this->e->add("Axis \"".$axe."\" doesn't exist for ID \"".$id."\"", "error");
     return false;
    }
    if($this->items[$axe]["axe"])
    {
     $this->e->add("Axis \"".$axe."\" is not a base axis for ID \"".$id."\"", "error");
     return false;
    }

    if($orientation===false)
    {
     if($this->items[$axe]["orientation"]=="horizontal")
      $orientation="vertical";
     else
      $orientation="horizontal";
    }

    if($this->items[$axe]["orientation"]==$orientation)
    {
     $this->e->add("Axis \"".$axe."\" has same orientation as ID \"".$id."\"", "error");
     return false;
    }
   }

   if($min!==false&&$max!==false&&$min>=$max)
   {
    $this->e->add("Maximum needs to be higher then minimum for ID \"".$id."\"", "error");
    return false;
   }

   $this->items[$id]=Array(
    "type" => "line",
    "gtype" => "axis",
    "order" => $this->order++,
    "orientation" => $orientation,
    "axe" => $axe,
    "rmin" => $min,
    "rmax" => $max,
    "rinterval" => false,
    "rplace" => false,
    "rlabels" => false,
    "indicator" => 0,
    "width" => 1,
    "margin" => 0,
    "color" => $this->_color("000000"),
    "coords" => Array(),
    "show" => true,
    "dateFormat" => $this->dateFormat,
    "display" => false,
    "displayAppend" => false,
    "fakeCoordsCounter" => 0,
   );

   return true;
  }

  function grid($axe, $interval=false)
  {
   if(!isset($this->items[$axe]))
   {
    $this->e->add("Axis \"".$axe."\" doesn't exist", "error");
    return false;
   }

   if($this->items[$axe]["gtype"]!="axis")
   {
    $this->e->add("ID \"".$id."\" isn't an axis", "error");
    return false;
   }

   $r=parent::grid("grid_".$axe, "0,0");

   $this->items["grid_".$axe]["gtype"]="grid";
   $this->items["grid_".$axe]["order"]=0;
   $this->items["grid_".$axe]["axe"]=$axe;
   $this->items["grid_".$axe]["interval"]=$interval;

   return $r;
  }

  function interval($axe, $interval=false)
  {
   if(!isset($this->items[$axe]))
   {
    $this->e->add("Axis \"".$axe."\" doesn't exist", "error");
    return false;
   }

   if($this->items[$axe]["gtype"]!="axis")
   {
    $this->e->add("ID \"".$id."\" isn't an axis", "error");
    return false;
   }

   $this->items[$axe]["rinterval"]=$interval;

   return true; 
  }

  function indicator($axe, $size=5)
  {
   if(!isset($this->items[$axe]))
   {
    $this->e->add("Axis \"".$axe."\" doesn't exist", "error");
    return false;
   }

   if($this->items[$axe]["gtype"]!="axis")
   {
    $this->e->add("ID \"".$id."\" isn't an axis", "error");
    return false;
   }

   $this->items[$axe]["indicator"]=$size;

   return true;
  }

  function place($axe, $place)
  {
   if(!isset($this->items[$axe]))
   {
    $this->e->add("Axis \"".$axe."\" doesn't exist", "error");
    return false;
   }

   if($place!=-1&&$place!=0&&$place!=1)
   {
    $this->e->add("Illegal place \"".$place."\"", "error");
    return false;
   }

   $this->items[$axe]["rplace"]=$place;
   return true;
  }

  function labels($axe)
  {
   if(!isset($this->items[$axe]))
   {
    $this->e->add("Axis \"".$axe."\" doesn't exist", "error");
    return false;
   }

   $args=func_get_args();
   array_shift($args);

   if(sizeof($args)==1&&is_array($args[0]))
    $args=$args[0];

   if(sizeof($args)==0)
    $this->items[$axe]["rlabels"]=true;
   else if(sizeof($args)==1&&($args[0]===true||$args[0]===false))
    $this->items[$axe]["rlabels"]=$args[0];
   else
    $this->items[$axe]["rlabels"]=$args;

   return true;
  }

  function tilt($axe=false)
  {
   if($axe===false)
   {
    foreach(array_keys($this->items) as $id) 
     if(isset($this->items[$id]["gtype"]))
      if($this->items[$id]["gtype"]=="axis"&&$this->items[$id]["axe"]===false) 
       $this->tilt($id);
   
    return true;
   }
   if(!isset($this->items[$axe]))
   {
    $this->e->add("Axis \"".$axe."\" doesn't exist", "error");
    return false;
   }

   if($this->items[$axe]["axe"]) //non base axe
    $axe=$this->items[$axe]["axe"];

   if($this->items[$axe]["orientation"]=="horizontal")
   {
    $this->items[$axe]["orientation"]="vertical"; //tilt base axe

    foreach(array_keys($this->items) as $id) //tilt related axis
     if(isset($this->items[$id]["gtype"]))
      if($this->items[$id]["gtype"]=="axis"&&$this->items[$id]["axe"]==$axe) 
       $this->items[$id]["orientation"]="horizontal"; 
   } 
   else
   {
    $this->items[$axe]["orientation"]="horizontal"; //tilt base axe

    foreach(array_keys($this->items) as $id) //tilt related axis
     if(isset($this->items[$id]["gtype"]))
      if($this->items[$id]["gtype"]=="axis"&&$this->items[$id]["axe"]==$axe) 
       $this->items[$id]["orientation"]="vertical"; 
   } 

   return true;
  }

  function dateFormat($axe=false, $dateFormat)
  {
   if($axe===false)
   {
    foreach(array_keys($this->items) as $id) 
     if(isset($this->items[$id]["gtype"]))
      if($this->items[$id]["gtype"]=="axis") 
       $this->dateFormat($id, $dateFormat);
   
    return true;
   }

   if(!isset($this->items[$axe]))
   {
    $this->e->add("Axis \"".$axe."\" doesn't exist", "error");
    return false;
   }

   $this->items[$axe]["display"]="date";
   $this->items[$axe]["dateFormat"]=$dateFormat;
   return true;
  }

  function display($axe=false, $display, $append=false)
  {
   if($axe===false)
   {
    foreach(array_keys($this->items) as $id) 
     if(isset($this->items[$id]["gtype"]))
      if($this->items[$id]["gtype"]=="axis") 
       $this->display($id, $display, $append);
   
    return true;
   }

   if(!isset($this->items[$axe]))
   {
    $this->e->add("Axis \"".$axe."\" doesn't exist", "error");
    return false;
   }

   if(!in_array($display, Array("date", "size", "humanreadable", "normal", false)))
   {
    $this->e->add("Unknown display method \"".$display."\" for axis \"".$axe."\"", "error");
    return false;
   }

   $this->items[$axe]["display"]=$display;
   $this->items[$axe]["displayAppend"]=$append;

   return true;
  }

  function margin($axe=false, $margin)
  {
   if($axe===false)
   {
    foreach(array_keys($this->items) as $id) 
     if(isset($this->items[$id]["gtype"]))
      if($this->items[$id]["gtype"]=="axis") 
       $this->margin($id, $margin);
   
    return true;
   }

   if(!isset($this->items[$axe]))
   {
    $this->e->add("Axis \"".$axe."\" doesn't exist", "error");
    return false;
   }

   $this->items[$axe]["margin"]=$margin;
   return true;
  }

  function line($id, $axe, $c1=false/*,$c2, $c3, $c4, etc*/)
  {
   if(isset($this->items[$id]))
   {
    $this->e->add("ID \"".$id."\" already exists", "error");
    return false;
   }

   if(!isset($this->items[$axe]))
   {
    $this->e->add("Axis \"".$axe."\" doesn't exist for ID \"".$id."\"", "error");
    return false;
   }

   if(!$this->items[$axe]["axe"])
   {
    $this->e->add("Axis \"".$axe."\" is a base axis for ID \"".$id."\"", "error");
    return false;
   }

   $args=func_get_args();
   array_shift($args);
   array_shift($args);

   if(is_array($args[0]))
    $args=$args[0];

   if(sizeof($args)<1)
   {
    $this->e->add("ID \"".$id."\" needs at least 1 coordinate", "error");
    return false;
   }

   $this->items[$id]=Array(
    "type" => "line",
    "gtype" => "line",
    "order" => $this->order++,
    "axe" => $axe,
    "width" => 1,
    "color" => $this->_color("000000"),
    "gcoords" => Array(),
    "show" => true,
   );

   $base=$this->items[$axe]["axe"];

   for($c=0;$c<sizeof($args);$c++)
   {
    if(!preg_match("/^(.*),([^,]+)$/", $args[$c], $tmp))
    {
     $this->e->add("Incorrect values for ID \"".$id."\"", "error");
     return false;
    }

    if(!is_numeric($tmp[1]))
    {
     if(($date=DT::dateConvertS($tmp[1], $this->items[$base]["dateFormat"]))!==false)
     {
      $c1=$date.",".$tmp[2];

      $this->items[$id]["gcoords"][]=$this->_coords($c1);

      if($this->items[$base]["display"]!="date")
       $this->items[$base]["display"]="date";
     }
     else
     {
      $text=$tmp[1];
      $c1=$this->items[$axe]["fakeCoordsCounter"].",".$tmp[2];

      $this->items[$axe]["fakeCoordsCounter"]++;
      $this->items[$id]["gcoords"][]=$this->_coords($c1);

      if($this->autoAddLineLabels)
       $this->text($id, $text, $c);
     }
    }
    else
     $this->items[$id]["gcoords"][]=$this->_coords($args[$c]);
   }

   return true;
  }

  function bar($id, $axe, $c1, $size=10)
  {
   if(isset($this->items[$id]))
   {
    $this->e->add("ID \"".$id."\" already exists", "error");
    return false;
   }

   if(!isset($this->items[$axe]))
   {
    $this->e->add("Axis \"".$axe."\" doesn't exist for ID \"".$id."\"", "error");
    return false;
   }

   if(!$this->items[$axe]["axe"])
   {
    $this->e->add("Axis \"".$axe."\" is a base axis for ID \"".$id."\"", "error");
    return false;
   }

   if(!preg_match("/^(.*),([^,]+)$/", $c1, $tmp))
   {
    $this->e->add("Incorrect values for ID \"".$id."\"", "error");
    return false;
   }

   $base=$this->items[$axe]["axe"];

   if(!is_numeric($tmp[1]))
   {
    if(($date=DT::dateConvertS($tmp[1], $this->items[$base]["dateFormat"]))!==false)
    {
     $text="";
     $c1=$date.",".$tmp[2];

     if($this->items[$base]["display"]!="date")
      $this->items[$base]["display"]="date";
    }
    else
    {
     $text=$tmp[1];
     $c1=$this->items[$axe]["fakeCoordsCounter"].",".$tmp[2];

     $this->items[$axe]["fakeCoordsCounter"]++;
    }
   }
   else
    $text="";

   $this->items[$id]=Array(
    "type" => "rectangle",
    "gtype" => "bar",
    "text" => $text,
    "order" => $this->order++,
    "axe" => $axe,
    "width" => 1,
    "size" => $size,
    "color" => $this->_color("FFFFFF"),
    "bColor" => $this->_color("000000"),
    "tColor" => $this->_color(false),
    "gcoords" => Array($this->_coords($c1)),
    "show" => true,
   );

   $this->textAlign($id, "center", "top");

   return true;
  }

  function render($width=800, $height=600)
  {
   if(!$this->size($width, $height))
    return false;

   foreach(array_keys($this->items) as $id)
    if(isset($this->items[$id]["gtype"]))
     if($this->items[$id]["gtype"]=="axis")
      $this->_determineAxe($id);

   foreach(array_keys($this->items) as $id)
    if(isset($this->items[$id]["gtype"]))
     if($this->items[$id]["gtype"]=="axis"&&!$this->items[$id]["axe"])
      $this->_redetermineBaseAxe($id);

   foreach(array_keys($this->items) as $id)
   {
    if(isset($this->items[$id]["gtype"]))
    {
     if($this->items[$id]["gtype"]=="axis")
     {
      $this->_determineAxeInterval($id);
      $this->_determineAxeLabels($id);
     }
     else if($this->items[$id]["gtype"]=="line")
      $this->_determineLineCoords($id);
     else if($this->items[$id]["gtype"]=="bar")
      $this->_determineBarCoords($id);
    }
   }

   foreach(array_keys($this->items) as $id)
   {
    if(isset($this->items[$id]["gtype"]))
    {
     switch($this->items[$id]["gtype"])
     {
      case "axis":
      {
       if($this->items[$id]["orientation"]=="vertical")
       {
        $coords=Array();
        $coords[]=Array($this->width*$this->items[$id]["place"], 0, 0);
        $coords[]=Array($this->width*$this->items[$id]["place"], $this->height, 0);
       }
       else
       {
        $coords=Array();
        $coords[]=Array(0, $this->height*(1-$this->items[$id]["place"]), 0);
        $coords[]=Array($this->width, $this->height*(1-$this->items[$id]["place"]), 0);
       }

       $this->items[$id]["coords"]=$coords;

       if($this->items[$id]["indicator"]||$this->items[$id]["labels"])
       {
        $interval=$this->items[$id]["interval"];
        $total=$this->items[$id]["max"]-$this->items[$id]["min"];
 
        if($total==0)
         $total=1;
 
        if($interval===0)
         break;

        if($this->items[$id]["orientation"]=="horizontal")
        {
         $x=0;
         $x=0;
         $labelStep=1;
         $step=$this->width*($interval/$total);

         if($this->items[$id]["labels"])
         {
          $avgLength=0;
          foreach($this->items[$id]["labels"] as $v)
           $avgLength+=strlen($v);

          $avgLength/=sizeof($this->items[$id]["labels"]);

          while($step<($this->fontWidth*$avgLength)) 
          {
           $step+=$this->width*($interval/$total);
           $labelStep++;
          }
         }

         for($c=0;$c<=$this->width;$c+=$step)
         {
          $this->items["_tmp_".$id."_".$c]=$this->items[$id];
 
          if($this->items[$id]["place"]==1)
           $this->items["_tmp_".$id."_".$c]["coords"]=Array(Array($c, ((1-$this->items[$id]["place"])*$this->height), 0), Array($c, ((1-$this->items[$id]["place"])*$this->height)+$this->items[$id]["indicator"], 0));
          else
           $this->items["_tmp_".$id."_".$c]["coords"]=Array(Array($c, ((1-$this->items[$id]["place"])*$this->height), 0), Array($c, ((1-$this->items[$id]["place"])*$this->height)-$this->items[$id]["indicator"], 0));

          if(isset($this->items[$id]["labels"][$x])&&$this->items[$id]["labels"][$x]!=="")
           $this->text("_tmp_".$id."_".$c, $this->items[$id]["labels"][$x]);

          $x+=$labelStep;
         }
        }
        else //vertical
        {
         $x=0;
         $labelStep=1;
         $step=$this->height*($interval/$total);

         if($this->items[$id]["labels"])
         {
          while($step<$this->fontHeight) 
          {
           $step*=2;
           $labelStep*=2;
          }
         }

         for($c=fmod($this->height, $step);$c<=$this->height;$c+=$step) //correct start with bottom left over
         {
          $this->items["_tmp_".$id."_".$c]=$this->items[$id];

          if($this->items[$id]["place"]==1)
           $this->items["_tmp_".$id."_".$c]["coords"]=Array(Array(($this->items[$id]["place"]*$this->width), $c, 0), Array(($this->items[$id]["place"]*$this->width)-$this->items[$id]["indicator"], $c, 0));
          else
           $this->items["_tmp_".$id."_".$c]["coords"]=Array(Array(($this->items[$id]["place"]*$this->width), $c, 0), Array(($this->items[$id]["place"]*$this->width)+$this->items[$id]["indicator"], $c, 0));

          if(isset($this->items[$id]["labels"][$x])&&$this->items[$id]["labels"][$x]!=="")
           $this->text("_tmp_".$id."_".$c, $this->items[$id]["labels"][$x]);

          $x+=$labelStep;
         }
        }
       }

       break;
      }
      case "grid":
      {
       $axe=$this->items[$id]["axe"];

       if($this->items[$id]["interval"])
        $interval=$this->items[$id]["interval"];
       else
        $interval=$this->items[$axe]["interval"];

       $total=$this->items[$axe]["max"]-$this->items[$axe]["min"];

       if($total==0)
        $total=1;

       if($interval===0)
        break;

       if($this->items[$axe]["orientation"]=="vertical")
        $coords=Array(0, $this->height*($interval/$total), 0);
       else
        $coords=Array($this->width*($interval/$total), 0, 0);

       $this->items[$id]["intervals"]=$coords;
        
       break;
      }
      default: break;
     }
    }
   }
   
   $r=parent::render();

   foreach(array_keys($this->items) as $id)
    if(preg_match("/^_tmp/", $id))
     unset($this->items[$id]);

   return $r;
  }

  function _determineLineCoords($id)
  {
   $coord=Array(0,0,0);
   $coords=Array();

   $axe=$this->items[$id]["axe"];
   $base=$this->items[$axe]["axe"];

   $bmin=$this->items[$base]["min"];
   $bmax=$this->items[$base]["max"];
   $bmiddle=$this->items[$base]["middle"];
   $btotal=$bmax-$bmin;

   $min=$this->items[$axe]["min"];
   $max=$this->items[$axe]["max"];
   $middle=$this->items[$axe]["middle"];
   $total=$max-$min;

   if($this->items[$base]["orientation"]=="horizontal")
   {
    $middle=1-$middle;
    foreach($this->items[$id]["gcoords"] as $c)
    {
     if($c[0]>=$bmin&&$c[0]<=$bmax)
     {
      if($c[1]>=$min&&$c[1]<=$max)
      {
       if($bmin>0)
        $coord[0]=$this->width*((($c[0]-$bmin)/$btotal)+$bmiddle);
       else if($bmax<0)
        $coord[0]=$this->width*((($c[0]-$bmax)/$btotal)+$bmiddle);
       else
        $coord[0]=$this->width*((($c[0])/$btotal)+$bmiddle);

       if($min>0)
        $coord[1]=$this->height*($middle-(($c[1]-$min)/$total));
       else if($max<0)
        $coord[1]=$this->height*($middle-(($c[1]-$max)/$total));
       else
        $coord[1]=$this->height*($middle-(($c[1])/$total));


       $coords[]=$coord;
      }
     }
    }
   }
   else
   {
    foreach($this->items[$id]["gcoords"] as $c)
    {
     if($c[0]>=$bmin&&$c[0]<=$bmax)
     {
      if($c[1]>=$min&&$c[1]<=$max)
      {
       if($min>0)
        $coord[0]=$this->width*((($c[1]-$min)/$total)+$middle);
       else if($max<0)
        $coord[0]=$this->width*((($c[1]-$max)/$total)+$middle);
       else
        $coord[0]=$this->width*((($c[1])/$total)+$middle);

       if($bmin>0)
        $coord[1]=$this->height*((1-$bmiddle)-(($c[0]-$bmin)/$btotal));
       else if($bmax<0)
        $coord[1]=$this->height*((1-$bmiddle)-(($c[0]-$bmax)/$btotal));
       else
        $coord[1]=$this->height*((1-$bmiddle)-(($c[0])/$btotal));

       $coords[]=$coord;
      }
     }
    }
   }

   $this->items[$id]["coords"]=$coords;
  }

  function _determineBarCoords($id)
  {
   $coord=Array(0,0,0);
   $coords=Array();

   $axe=$this->items[$id]["axe"];
   $base=$this->items[$axe]["axe"];

   $bmin=$this->items[$base]["min"];
   $bmax=$this->items[$base]["max"];
   $bmiddle=$this->items[$base]["middle"];
   $btotal=$bmax-$bmin;

   $min=$this->items[$axe]["min"];
   $max=$this->items[$axe]["max"];
   $middle=$this->items[$axe]["middle"];
   $total=$max-$min;

   if($this->items[$base]["orientation"]=="horizontal")
   {
    $middle=1-$middle;
    foreach($this->items[$id]["gcoords"] as $c)
    {
     if($c[0]>=$bmin&&$c[0]<=$bmax)
     {
      if($c[1]>=$min&&$c[1]<=$max)
      {
       if($bmin>0)
        $coord[0]=($this->width*((($c[0]-$bmin)/$btotal)+$bmiddle))-($this->items[$id]["size"]/2);
       else if($bmax<0)
        $coord[0]=($this->width*((($c[0]-$bmax)/$btotal)+$bmiddle))-($this->items[$id]["size"]/2);
       else
        $coord[0]=($this->width*((($c[0])/$btotal)+$bmiddle))-($this->items[$id]["size"]/2);

       if($min>0)
        $coord[1]=$this->height*($middle-(($c[1]-$min)/$total));
       else if($max<0)
        $coord[1]=$this->height*($middle-(($c[1]-$max)/$total));
       else
        $coord[1]=$this->height*($middle-(($c[1])/$total));

       $coords[]=$coord;
 
       $coord[0]+=$this->items[$id]["size"];
       $coord[1]=$this->height*$middle;
       $coords[]=$coord;
      }
     }
    }
   }
   else
   {
    foreach($this->items[$id]["gcoords"] as $c)
    {
     if($c[0]>=$bmin&&$c[0]<=$bmax)
     {
      if($c[1]>=$min&&$c[1]<=$max)
      {
       if($min>0)
        $coord[0]=$this->width*((($c[1]-$min)/$total)+$middle);
       else if($max<0)
        $coord[0]=$this->width*((($c[1]-$max)/$total)+$middle);
       else
        $coord[0]=$this->width*((($c[1])/$total)+$middle);

       if($bmin>0)
        $coord[1]=($this->height*((1-$bmiddle)-(($c[0]-$bmin)/$btotal)))-($this->items[$id]["size"]/2);
       else if($bmax<0)
        $coord[1]=($this->height*((1-$bmiddle)-(($c[0]-$bmax)/$btotal)))-($this->items[$id]["size"]/2);
       else
        $coord[1]=($this->height*((1-$bmiddle)-(($c[0])/$btotal)))-($this->items[$id]["size"]/2);

       $coords[]=$coord;
 
       $coord[0]=$this->width*$middle;
       $coord[1]+=$this->items[$id]["size"];
       $coords[]=$coord;
      }
     }
    }
   }

   if($coords)
    $this->show($id);
   else
    $this->hide($id);

   $this->items[$id]["coords"]=$coords;
  }

  function _determineAxe($axe)
  {
   //get extremes
   if($this->items[$axe]["orientation"]=="horizontal")
   {
    if($this->items[$axe]["rmin"]===false||$this->items[$axe]["rmax"]===false)
    {
     $xmin=false;
     $xmax=false;

     foreach(array_keys($this->items) as $id)
     {
      if(isset($this->items[$id]["gcoords"]))
      {
       if($this->items[$axe]["axe"]===false&&$this->items[$this->items[$id]["axe"]]["axe"]==$axe) //base axis, use first coord 
       {
        foreach($this->items[$id]["gcoords"] as $c)
        {
         if($xmin===false||$c[0]<$xmin)
          $xmin=$c[0];
         if($xmax===false||$c[0]>$xmax)
          $xmax=$c[0];
        }
       }
       else if($this->items[$id]["axe"]==$axe) //values
       {
        foreach($this->items[$id]["gcoords"] as $c)
        {
         if($xmin===false||$c[1]<$xmin)
          $xmin=$c[1];
         if($xmax===false||$c[1]>$xmax)
          $xmax=$c[1];
        }
       }
      }
     }

     if($this->items[$axe]["rmin"]===false)
      $this->items[$axe]["min"]=$xmin;
     else
      $this->items[$axe]["min"]=$this->items[$axe]["rmin"];

     if($this->items[$axe]["rmax"]===false)
      $this->items[$axe]["max"]=$xmax;
     else
      $this->items[$axe]["max"]=$this->items[$axe]["rmax"];
    }
    else
    {
     $this->items[$axe]["min"]=$this->items[$axe]["rmin"];
     $this->items[$axe]["max"]=$this->items[$axe]["rmax"];
    }
   }
   else
   {
    if($this->items[$axe]["rmin"]===false||$this->items[$axe]["rmax"]===false)
    {
     $ymin=false;
     $ymax=false;

     foreach(array_keys($this->items) as $id)
     {
      if(isset($this->items[$id]["gcoords"]))
      {
       if($this->items[$axe]["axe"]===false&&$this->items[$this->items[$id]["axe"]]["axe"]==$axe) //base axis, use first coord 
       {
        foreach($this->items[$id]["gcoords"] as $c)
        {
         if($ymin===false||$c[0]<$ymin)
          $ymin=$c[0];
         if($ymax===false||$c[0]>$ymax)
          $ymax=$c[0];
        }
       }
       else if($this->items[$id]["axe"]==$axe) //values
       {
        foreach($this->items[$id]["gcoords"] as $c)
        {
         if($ymin===false||$c[1]<$ymin)
          $ymin=$c[1];
         if($ymax===false||$c[1]>$ymax)
          $ymax=$c[1];
        }
       }
      }
     }

     if($this->items[$axe]["rmin"]===false)
      $this->items[$axe]["min"]=$ymin;
     else
      $this->items[$axe]["min"]=$this->items[$axe]["rmin"];

     if($this->items[$axe]["rmax"]===false)
      $this->items[$axe]["max"]=$ymax;
     else
      $this->items[$axe]["max"]=$this->items[$axe]["rmax"];
    }
    else
    {
     $this->items[$axe]["min"]=$this->items[$axe]["rmin"];
     $this->items[$axe]["max"]=$this->items[$axe]["rmax"];
    }
   }

   if($this->items[$axe]["margin"])
   {
    $margin=($this->items[$axe]["max"]-$this->items[$axe]["min"])*$this->items[$axe]["margin"];

    if($this->items[$axe]["rmin"]===false&&$this->items[$axe]["min"]!=0)
     $this->items[$axe]["min"]-=$margin;
    if($this->items[$axe]["rmax"]===false&&$this->items[$axe]["max"]!=0)
     $this->items[$axe]["max"]+=$margin;
   }

   if($this->items[$axe]["min"]==$this->items[$axe]["max"])
    $this->items[$axe]["max"]++;

   //determine middle of axe
   if($this->items[$axe]["min"]>=0)
    $this->items[$axe]["middle"]=0;
   else if($this->items[$axe]["max"]<=0)
    $this->items[$axe]["middle"]=1;
   else
    $this->items[$axe]["middle"]=(-$this->items[$axe]["min"])/($this->items[$axe]["max"]-$this->items[$axe]["min"]);
  }

  function _redetermineBaseAxe($axe)
  {
   $axis=Array();

   //get corresponding axis
   foreach(array_keys($this->items) as $id)
    if(isset($this->items[$id]["gtype"]))
     if($this->items[$id]["gtype"]=="axis"&&$this->items[$id]["axe"]==$axe)
      $axis[]=$id;

   $middles=Array();
   //get different middle postions
   foreach($axis as $id)
   {
    if($this->items[$id]["middle"]>0&&$this->items[$id]["middle"]<1)
    {
     if(!isset($middles[(string)$this->items[$id]["middle"]]))
      $middles[(string)$this->items[$id]["middle"]]=Array();

     $middles[(string)$this->items[$id]["middle"]][]=$id;
    }
   }

   $middle=false;
   //determine most prominant middle position
   foreach(array_keys($middles) as $pos)
   {
    if($middle===false)
     $middle=$pos;
    else if(sizeof($middles[$pos])>sizeof($middles[$middle]))
     $middle=$pos;
   }

   if($middle)
   {
    //sync axis
    foreach($axis as $id)
    {
     if($this->items[$id]["middle"]<$middle) //upscale min
     {
      $newtotal=$this->items[$id]["max"]/(1-$middle);
      $this->items[$id]["min"]=-1*($newtotal*$middle);
      $this->items[$id]["middle"]=$middle;
     }
     else if($this->items[$id]["middle"]>$middle) //upscale max 
     {
      $newtotal=-1*($this->items[$id]["min"]/$middle);
      $this->items[$id]["max"]=$newtotal*(1-$middle);
      $this->items[$id]["middle"]=$middle;
     }
    }
   }

   //base axe place
   if($this->items[$axe]["rplace"]===false) //auto determine, always middle
    $this->items[$axe]["place"]=$middle; 
   else if($this->items[$id]["rplace"]==0) //middle position
    $this->items[$id]["place"]=$middle;
   else if($this->items[$axe]["rplace"]==-1) //left/top
    $this->items[$id]["place"]=0; 
   else //should be right/bottom
    $this->items[$axe]["place"]=$this->items[$axe]["rplace"]; 

   $takenPlaces=Array();
   //axe place
   foreach($axis as $id)
   {
    if($this->items[$id]["rplace"]!==false)
     $place=$this->items[$id]["rplace"];
    else //auto determine
    {
     if(sizeof($axis)==1)
      $place=0;
     else if(sizeof($axis)==2)
     {
      if(!in_array(-1, $takenPlaces))
       $place=-1;
      else if(!in_array(1, $takenPlaces))
       $place=1;
     }
     else
     {
      if(!in_array(-1, $takenPlaces))
       $place=-1;
      else if(!in_array(0, $takenPlaces))
       $place=0;
      else if(!in_array(1, $takenPlaces))
       $place=1;
      else
       $place=0;
     }
    }
    
    if($place==-1) //left/top
     $this->items[$id]["place"]=0; 
    else if($place==0) //middle position, depending on base axe middle
     $this->items[$id]["place"]=$this->items[$axe]["middle"];
    else //should be right/bottom
     $this->items[$id]["place"]=$place; 

    $takenPlaces[]=$place;
   }

   $this->items[$axe]["place"]=$middle;
  }

  function _determineAxeInterval($axe)
  {
   if($this->items[$axe]["rinterval"])
    $this->items[$axe]["interval"]=$this->items[$axe]["rinterval"];
   else
   {
    $total=$this->items[$axe]["max"]-$this->items[$axe]["min"];

    if($total==0)
     $total=1;

    if(($items=sizeof($this->items[$axe]["rlabels"]))>1)
     $items;
    else
    {
     $items=0;
     if($this->items[$axe]["axe"]===false) //base
     {
      foreach(array_keys($this->items) as $id)
      {
       if(isset($this->items[$id]["gtype"])&&$this->items[$id]["gtype"]=="axis"&&$axe==$this->items[$id]["axe"])
       {
        $tmp=0;
        foreach(array_keys($this->items) as $cid)
         if(isset($this->items[$cid]["gcoords"])&&$this->items[$cid]["axe"]==$id)
          $tmp+=sizeof($this->items[$cid]["gcoords"]);

        if($tmp>$items)
         $items=$tmp;
       }
      }
     }
     else
     {
      foreach(array_keys($this->items) as $cid)
       if(isset($this->items[$cid]["gcoords"])&&$this->items[$cid]["axe"]==$axe)
        $items+=sizeof($this->items[$cid]["gcoords"]);
     }
    }

    if($items>1)
     $interval=$total/($items-1);
    else
     $interval=$total/2;

    $this->items[$axe]["interval"]=$interval;
   }
  }

  function _determineAxeLabels($axe)
  {
   if($this->items[$axe]["rlabels"]===false||is_array($this->items[$axe]["rlabels"])) //no labels or static labels
    $this->items[$axe]["labels"]=$this->items[$axe]["rlabels"];
   else //generate labels
   {
    $this->items[$axe]["labels"]=Array();
    $interval=$this->items[$axe]["interval"];
 
    for($c=$this->items[$axe]["min"];$c<=$this->items[$axe]["max"];$c+=$this->items[$axe]["interval"])
    {
     if($this->items[$axe]["display"]=="date")
      $this->items[$axe]["labels"][]=date($this->items[$axe]["dateFormat"], round($c)).($this->items[$axe]["displayAppend"]?$this->items[$axe]["displayAppend"]:"");
     else if($this->items[$axe]["display"]=="size")
      $this->items[$axe]["labels"][]=Mutilate::humanReadableFileSizeS($c, 1).($this->items[$axe]["displayAppend"]?$this->items[$axe]["displayAppend"]:"");
     else if($this->items[$axe]["display"]=="humanreadable")
      $this->items[$axe]["labels"][]=Mutilate::humanReadableNumberS($c, 1).($this->items[$axe]["displayAppend"]?$this->items[$axe]["displayAppend"]:"");
     else
      $this->items[$axe]["labels"][]=round($c, 5).($this->items[$axe]["displayAppend"]?$this->items[$axe]["displayAppend"]:"");
    }
   }

   if($this->items[$axe]["labels"]!==false) //axe has labels
   {
    if($this->items[$axe]["orientation"]=="vertical") //reverse labels for vertical axis
     $this->items[$axe]["labels"]=array_reverse($this->items[$axe]["labels"]);

    //remove labels that overlap with axe title
    $label=false;
    foreach(array_keys($this->items) as $id) //find axe title
    {
     if($this->items[$id]["type"]=="label"&&$this->items[$id]["id"]==$axe) //axe title found
     {
      $label=true;

      if($this->items[$id]["coord"]==0) //title at left or top
      { 
       if($this->items[$axe]["orientation"]=="vertical") //vertical axe
        $this->items[$axe]["labels"][0]=""; //remove top entry
       else //horizontal axe, remove first couple of entries
       {
        //determine label width
        if($this->items[$id]["width"])
         $labelWidth=$this->items[$id]["width"];
        else
         $labelWidth=strlen($this->items[$id]["text"])*$this->fontWidth;

        $entrieSize=$this->width/sizeof($this->items[$axe]["labels"]);
        $entries=floor($labelWidth/$entrieSize); //width / entry size = number of entries to be removed

        for($c=0;$c<$entries;$c++)
         $this->items[$axe]["labels"][$c]=""; //remove first $entries number of entries from axe
       }
      }
      else //title at right or bottom
      {
       if($this->items[$axe]["orientation"]=="vertical") //vertical axe
        $this->items[$axe]["labels"][sizeof($this->items[$axe]["labels"])-1]=""; //remove bottom entry
       else //horizontal axe
        $this->items[$axe]["labels"][sizeof($this->items[$axe]["labels"])-1]=""; //remove right entry
      }

      break;
     }
    }

    if($base=$this->items[$axe]["axe"]) //non base axis
    {
     if($this->items[$axe]["orientation"]=="vertical") //vertical non base axe
     {
      //remove labels that overlap with labels from base axe
      if($this->items[$base]["labels"]!==false) //base axe has labels
      {
       if( //axis both in the middle (crossing)
        $this->items[$axe]["place"]>0&&$this->items[$axe]["place"]<1&&
        $this->items[$base]["place"]>0&&$this->items[$base]["place"]<1
       )
       {
        //remove entry contain "0" (crossing) value from base axis
        if(($entry=array_search("0", $this->items[$base]["labels"], true))!==false)
         $this->items[$base]["labels"][$entry]="";
       }

       $this->items[$axe]["labels"][sizeof($this->items[$axe]["labels"])-1]=""; //remove right/bottom entry
      }
      else //base axe has title?
      {
       foreach(array_keys($this->items) as $id) //find base axe title
       {
        if($this->items[$id]["type"]=="label"&&$this->items[$id]["id"]==$base) //base axe title found
        {
         $this->items[$axe]["labels"][sizeof($this->items[$axe]["labels"])-1]=""; //remove right/bottom entry
         break;
        }
       }
      }


      //remove labels from base axe that overlap with axe title
      foreach(array_keys($this->items) as $id) //find axe title
      {
       if($this->items[$id]["type"]=="label"&&$this->items[$id]["id"]==$axe) //axe title found
       {
        if($this->items[$id]["coord"]!=0) //title at bottom
        {
         if($this->items[$id]["width"])
          $labelWidth=$this->items[$id]["width"];
         else
          $labelWidth=strlen($this->items[$id]["text"])*$this->fontWidth;

         $entrieSize=$this->width/sizeof($this->items[$base]["labels"]);
         $entries=floor($labelWidth/$entrieSize); //width / entry size = number of entries to be removed

         for($c=0;$c<$entries;$c++)
          $this->items[$base]["labels"][$c]=""; //remove first $entries number of entries from base axe
        }

        break;
       }
      }
     }
     else //horizontal axe
     {
      //remove labels from base axe that overlap with axe title
      foreach(array_keys($this->items) as $id) //find axe title
      {
       if($this->items[$id]["type"]=="label"&&$this->items[$id]["id"]==$axe) //axe title found
       {
        if($this->items[$id]["coord"]==0) //title at left 
         $this->items[$base]["labels"][sizeof($this->items[$base]["labels"])-1]=""; //remove bottom entry

        break;
       }
      }


      //remove labels from axe that overlap with base axe title
      foreach(array_keys($this->items) as $id) //find axe title
      {
       if($this->items[$id]["type"]=="label"&&$this->items[$id]["id"]==$base) //axe title found
       {
        if($this->items[$id]["coord"]==0) //title at top
        {
         if($this->items[$id]["width"])
          $labelWidth=$this->items[$id]["width"];
         else
          $labelWidth=strlen($this->items[$id]["text"])*$this->fontWidth;

         $entrieSize=$this->width/sizeof($this->items[$axe]["labels"]);
         $entries=ceil($labelWidth/$entrieSize); //width / entry size = number of entries to be removed

         for($c=0;$c<$entries;$c++)
          $this->items[$axe]["labels"][$c]=""; //remove first $entries number of entries from base axe
        }

        break;
       }
      }
     }
    }
   }


  }

 }

?>
